/*
 * Copyright 2014 Open Networking Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ctpd.closfwd;

import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.Iterator;

import org.apache.karaf.shell.commands.Argument;
import org.apache.karaf.shell.commands.Command;
import org.onosproject.cli.AbstractShellCommand;
import org.onosproject.net.HostId;

import org.ctpd.closfwd.api.ClosDeviceService;
import org.ctpd.closfwd.api.Endpoint;
import org.ctpd.closfwd.api.OltControlEndpoint;
import org.ctpd.closfwd.api.VpdcHostEndpoint;
import org.onosproject.core.ApplicationId;

import org.onlab.packet.MacAddress;
import org.onlab.packet.VlanId;
import org.onosproject.net.DeviceId;
import org.onosproject.net.PortNumber;

/**
 * A demo service that lists the endpoints for which intents are installed.
 */
@Command(scope = "onos", name = "recreate-intents-clos",
        description = "Reinstall endpoint UUID given by argument in clos")
public class CtpdRecreateIntentsClosCommand extends AbstractShellCommand {


    @Argument(index = 0, name = "uuid", description = "Recreate endpoint with UUID given. Be careful with dependencies between some EPs",
    required = false, multiValued = false)
    private String uuid = null;


    // reference to our service
    private ClosDeviceService service;
    @Override
    protected void execute() {
        service = get(ClosDeviceService.class);

        if(uuid==null){
            
            print("Recreating Endpoints depending on intents:");

            Iterator<UUID> uuids = service.getRegisterUUIDs().iterator();
            while(uuids.hasNext())
            {
                UUID uuid = (UUID) uuids.next();
                Endpoint endpoint = service.getEndpoint(uuid);
                ApplicationId appId = service.getApplicationFlowId(endpoint);
                if(endpoint instanceof OltControlEndpoint || endpoint instanceof VpdcHostEndpoint){
                    service.recreateEndpoint(endpoint);
                    print(String.format("UUID:%40s  Vlan:%8s  Mac:%20s  Type:%30s  ApplicationId:%6s",uuid.toString(), endpoint.getVlan().toString(), endpoint.getMac(),
                    endpoint.getClass().getName().substring(endpoint.getClass().getName().lastIndexOf('.') + 1), appId.id()));
                }

            }

        }

    }
}