/*
 * Copyright 2016 Open Networking Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Application developed by Elisa Rojas Sanchez
 */
package org.ctpd.olt.api;

import org.onosproject.net.DeviceId;
import org.onosproject.net.PortNumber;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonSubTypes;
import com.fasterxml.jackson.annotation.JsonSubTypes.Type;
import com.fasterxml.jackson.annotation.JsonTypeInfo;
import com.google.common.base.Objects;
import static java.util.Optional.ofNullable;



public class OltDevice {
	public DeviceId deviceId;
	//public PortNumber firstInputPort;
	//public PortNumber lastInputPort;
	public PortNumber uplink;

	@JsonCreator
	public OltDevice(@JsonProperty("device") String deviceId,
			@JsonProperty("uplink") long uplink) {
		this.uplink = PortNumber.portNumber(uplink);
		this.deviceId = DeviceId.deviceId(deviceId);
		
	}

	/* Get methods */
	public DeviceId getDeviceId() {
		return this.deviceId;
	}

	public PortNumber getUplink() {
		return this.uplink;
	}
	/* End get methods */


	/* Serialization functions */
	@JsonProperty("device")
	public String getDeviceIdSer() {
		return this.deviceId.toString();
	}

	@JsonProperty("uplink")
	public String getUplinkSer() {
		return this.uplink.toString();
	}
	/* End serialization functions */

	@Override
	public String toString() {
		return String.format("device:%s, uplink:%s",
				ofNullable(this.deviceId.toString()).orElse("NULL"),
				ofNullable(this.uplink.toString()).orElse("NULL"));
	}

	@Override
	public int hashCode() {
		return Objects.hashCode(this.deviceId.hashCode(),
				this.uplink.hashCode());
	}

	@Override
	public boolean equals(Object other) {
		if (!(other instanceof OltDevice))
			return false;
		OltDevice otherService = (OltDevice) other;
		if (otherService.deviceId == this.deviceId
				&& otherService.uplink.equals(this.uplink)) {
			return true;
		}
		return false;
	}
}
