/*
 * Copyright 2015 Open Networking Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ctpd.olt.api;

import org.onlab.packet.VlanId;
import org.onlab.packet.Ip4Address;
import org.onlab.packet.IpAddress;
import org.onlab.packet.MacAddress;
import org.onlab.packet.IpPrefix;
import org.onosproject.event.ListenerService;
import org.onosproject.net.ConnectPoint;
import org.onosproject.net.DeviceId;
import org.onosproject.net.PortNumber;

import java.util.Collection;
import java.util.Map;

/**
 * Service for interacting with an access device (OLT).
 */
public interface OltDeviceService {
        // extends ListenerService<AccessDeviceEvent, AccessDeviceListener> {

    /**
     * Provisions connectivity for a subscriber on an access device.
     *
     * @param port subscriber's connection point
     * @param ipPrefix subscribers ip prefix
     * @param vlan VLAN ID to provision for subscriber
     * @param bandwidth bandwidth limit associated to this subscriber
     */

    String addSubscriber(ConnectPoint port,  VlanId inVlan, int bandwidth);

    /**
     * Provisions connectivity for a subscriber on an access device.
     *
     * @param deviceId olt deviceId
     * @param inPort subscriber's port
     */
    //String removeSubscriberCtpd(DeviceId deviceId, PortNumber inPort);
    String removeSubscriber(DeviceId deviceId, PortNumber inPort);

    /**
     *  Adds a device to the list of OLT devices
     *
     * @param device olt deviceId
     * @return String error or success
     */
    public String addDevice(OltDevice device);

    /**
     *  Remove a device to the list of OLT devices
     *
     * @param device olt deviceId
     * @return String error or success
     */
    String removeDevice(DeviceId device);

    /**
     *  Purge OLT, to initial state
     */
    public void purgeOLT();

    /**
     *  Get all data stored for all subscribers on all OLT's
     */
    public String getOLTSubscribers();

}
