/*
 * Copyright 2016 Open Networking Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
package org.ctpd.olt.api;

import org.onosproject.net.DeviceId;
import org.onosproject.net.PortNumber;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonSubTypes;
import com.fasterxml.jackson.annotation.JsonSubTypes.Type;
import com.fasterxml.jackson.annotation.JsonTypeInfo;
import org.onlab.packet.VlanId;

import com.google.common.base.Objects;
import static java.util.Optional.ofNullable;



public class OltVlanPort {
	public PortNumber portId;
	public int bandwidth;
	public VlanId vlanId;

	@JsonCreator
	public OltVlanPort(@JsonProperty("port") String portId,
			@JsonProperty("bandwidth") String bandwidth, @JsonProperty("vlan") String vlan) {

		this.portId = PortNumber.portNumber(portId);
		this.bandwidth = Integer.parseInt(bandwidth);
	  	if (vlan.equals("-1"))
            this.vlanId = VlanId.vlanId(VlanId.UNTAGGED);
        else
        	this.vlanId = VlanId.vlanId(vlan);
	}

	/* Get methods */


	public PortNumber getPort() {
		return this.portId;
	}

	public int getBandwidth() {
		return this.bandwidth;
	}

	public VlanId getVlan() {
		return this.vlanId;
	}


	/* Serialization functions */
	@JsonProperty("port")
	public String getPortSer() {
		return this.portId.toString();
	}

	@JsonProperty("bandwidth")
	public String getBandwidthSer() {
		return String.valueOf(this.bandwidth);
	}

	@JsonProperty("vlan")
	public String getVlanSer() {
		return this.vlanId.toString();
	}

	/* End serialization functions */

	@Override
	public String toString() {
		return String.format("port:%s, bandwidth:%s, vlan:%s",
				ofNullable(this.portId.toString()).orElse("NULL"),
				ofNullable(String.valueOf(this.bandwidth)).orElse("NULL"), ofNullable(this.vlanId.toString()).orElse("NULL") );
	}

	@Override
	public int hashCode() {
		return Objects.hashCode(this.portId.hashCode(),
				String.valueOf(this.bandwidth).hashCode(), this.vlanId.hashCode());
	}

	@Override
	public boolean equals(Object other) {
		if (!(other instanceof OltVlanPort))
			return false;
		OltVlanPort otherPort = (OltVlanPort) other;
		if (otherPort.portId ==this.portId && otherPort.bandwidth == this.bandwidth
				&& otherPort.vlanId == this.vlanId) {
			return true;
		}
		return false;
	}
}
