/*
 * Copyright 2014 Open Networking Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ctpd.olt;

import org.apache.karaf.shell.commands.Argument;
import org.apache.karaf.shell.commands.Command;
import org.onosproject.cli.AbstractShellCommand;

import org.ctpd.olt.api.OltDeviceService;

import org.onlab.packet.VlanId;
import org.onosproject.net.PortNumber;
import org.onosproject.net.DeviceId;


/**
 * A demo service that lists the endpoints for which intents are installed.
 */
@Command(scope = "onos", name = "remove-olt-subscriber",
        description = "Remove OLT Subscriber by Port and Vlan")
public class CtpdOLTRemoveSuscribersCommand extends AbstractShellCommand {
	@Argument(index = 0, name = "device", description = "Device Id of Subscriber",
            required = true, multiValued = false)
    private String device = null;

	@Argument(index = 1, name = "portnumber", description = "Port Number of Subscriber",
            required = true, multiValued = false)
    private String port = null;

    @Argument(index = 2, name = "vlan", description = "Vlan Id of Subscriber",
            required = true, multiValued = false)
    private String vlan = null;

    // reference to our service
    private OltDeviceService service;

    @Override
    protected void execute() {
    	DeviceId deviceId = DeviceId.deviceId(device);
    	PortNumber portNumber = PortNumber.portNumber(port);
    	VlanId vlanId = VlanId.vlanId(vlan);
        print(get(OltDeviceService.class).removeSubscriber(deviceId, portNumber));
    }
}